package com.ichi2.anki;

import android.content.Context;
import android.content.res.Resources;
import android.graphics.Typeface;
import android.util.Log;
import android.widget.Toast;

import com.ichi2.libanki.Utils;

import java.io.File;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class AnkiFont {
    private String mName;
    private String mFamily;
    private List<String> mAttributes;
    private String mPath;
    private static final String fAssetPathPrefix = "/android_asset/fonts/";
    private static Set<String> corruptFonts = new HashSet<String>();

    private AnkiFont(String name, String family, List<String> attributes, String path) {
        mName = name;
        mFamily = family;
        mAttributes = attributes;
        mPath = path;
    }
    
    /**
     * Factory for AnkiFont creation.
     * Creates a typeface wrapper from a font file representing.
     * @param ctx Activity context, needed to access assets
     * @param path Path to typeface file, needed when this is a custom font.
     * @param fromAssets True if the font is to be found in assets of application
     * @return A new AnkiFont object or null if the file can't be interpreted as typeface.
     */
    public static AnkiFont createAnkiFont(Context ctx, String path, boolean fromAssets) {
        File fontfile = new File(path);
        String name = Utils.removeExtension(fontfile.getName());
        String family = name;
        List<String> attributes = new ArrayList<String>();

        if (fromAssets) {
            path = fAssetPathPrefix.concat(fontfile.getName());
        }
        Typeface tf = getTypeface(ctx, path);
        if (tf == null) {
            // unable to create typeface
            return null;
        }
        if (tf.isBold() || name.toLowerCase().contains("bold")) {
            attributes.add("font-weight: bolder;");
            family = family.replaceFirst("(?i)-?Bold", "");
        } else if (name.toLowerCase().contains("light")) {
            attributes.add("font-weight: lighter;");
            family = family.replaceFirst("(?i)-?Light", "");
        } else {
            attributes.add("font-weight: normal;");
        }
        if (tf.isItalic() || name.toLowerCase().contains("italic")) {
            attributes.add("font-style: italic;");
            family = family.replaceFirst("(?i)-?Italic", "");
        } else if (name.toLowerCase().contains("oblique")) {
            attributes.add("font-style: oblique;");
            family = family.replaceFirst("(?i)-?Oblique", "");
        } else {
            attributes.add("font-style: normal;");
        }
        if (name.toLowerCase().contains("condensed") || name.toLowerCase().contains("narrow")) {
            attributes.add("font-stretch: condensed;");
            family = family.replaceFirst("(?i)-?Condensed", "");
            family = family.replaceFirst("(?i)-?Narrow(er)?", "");
        } else if (name.toLowerCase().contains("expanded") || name.toLowerCase().contains("wide")) {
            attributes.add("font-stretch: expanded;");
            family = family.replaceFirst("(?i)-?Expanded", "");
            family = family.replaceFirst("(?i)-?Wide(r)?", "");
        } else {
            attributes.add("font-stretch: normal;");
        }
        family = family.replaceFirst("(?i)-?Regular", "");
        return new AnkiFont(name, family, attributes, path);
    }
    
    public String getDeclaration() {
        StringBuilder sb = new StringBuilder("@font-face {");
        sb.append(getCSS()).append(" src: url(\"file://").append(mPath).append("\");}");
        return sb.toString();
    }
    public String getCSS() {
        StringBuilder sb = new StringBuilder("font-family: \"").append(mFamily).append("\";");
        for (String attr : mAttributes) {
            sb.append(" ").append(attr);
        }
        return sb.toString();
    }
    public String getName() {
        return mName;
    }
    public String getPath() {
        return mPath;
    }
    public static Typeface getTypeface(Context ctx, String path) {
        try {
            if (path.startsWith(fAssetPathPrefix)) {
                return Typeface.createFromAsset(ctx.getAssets(), path.replaceFirst("/android_asset/", ""));
            } else {
                return Typeface.createFromFile(path);
            }
        } catch (RuntimeException e) {
            Log.w(AnkiDroidApp.TAG, "AnkiFont.getTypeface: " + e.getMessage() + " - File: " + path);
            if (!corruptFonts.contains(path)) {
                // Show warning toast
                String name = new File(path).getName();
                Resources res = AnkiDroidApp.getAppResources();
                Toast toast = Toast.makeText(ctx, res.getString(R.string.corrupt_font, name), Toast.LENGTH_LONG);
                toast.show();
                // Don't warn again in this session
                corruptFonts.add(path);
            }
            return null;
        }
    }
}
